/************************************************************************/
/*                                                                      */
/*             ,     ..        */
/*                   ,   ,   .               */
/*                                                                      */
/************************************************************************/

#include <htc.h>
#include "Include\PIC_Common.h"
#include "sysnet.h"
#include "ir.h"
#include "eeprom.h"

/**********************************************************
Project: Ringer
Author:  Sergey B. Voinov
Created: 11/06/2009
MCU:     PIC16F877 QFP 44

Timer usage:
	IR receiver         - timer 0
	RS485 receiver      - timer 1
	Sample rate control - timer 2

Interrupts usage:
	Internal - timer 0 (IR receiver timer overflow)
	External - RB0 (connected to TSOP1736) - rising and falling edges

Communications:
	RS485 - not yet enabled (HW bug in RX pin junction between 
			RS232 and RS485 - currently hardwired to RS232).
	RS232 - uses custom baud rate of 1800. Since EEPROM write time
			is 5ms (for single byte - this method is used for now),
			and 2400 baud rate is too high for it. 1200 baud rate
			cannot be implemented in MCU's UART since it overflows
			baud rate register. So 1800 is optimal. Contemporary
			motherboards support such exotic baud rates, so there
			is no problem except slow transfer rate and non-efficient
			EEPROM write cycle usage (should rather use page-write).
	
Warning: refactoring doesn't work fine for microcontroller projects.
Too deeply nesting routines produce compiler warning 
"possible stack overflow when calling function". It really messes all
up in some circumstances, so it is better to reduce routine nesting
if such warning appears.

TODO: should define F_OSC at project level, not in a global include file.
For now, some places in code include its own definition of F_OSC, which
is inconvenient and error-prone. All code that uses timers is impacted.

-----------------------------------------------------------
Pin  Dir  Description
-----------------------------------------------------------
RA0  I    Optocoupler output
RA1  O    RS-232 interface enable (high)
RA2  O    RS-485 interface enable (high)
RA3  O    DAC A/B pin
RA4  NC
RA5  O    DAC /PD pin
RA6  -    unused
RA7  -    unused

RB0  I    Interrupt from IR detector
RB1  O    EEPROM write protect enable
RB2  NC
RB3  NC
RB4  NC
RB5  NC
RB6  -    ICSP clock
RB7  -    ICSP data

RC0  O    RS-485 receiver enable (low)
RC1  O    RS-485 driver enable (high)
RC2  O    DAC /CLR pin
RC3  -    I2C clock
RC4  -    I2C data
RC5  NC
RC6  -    USART TX
RC7  -    USART RX

RD0  O    DAC D0
RD1  O    DAC D1
RD2  O    DAC D2
RD3  O    DAC D3
RD4  O    DAC D4
RD5  O    DAC D5
RD6  O    DAC D6
RD7  O    DAC D7

RE0  O    DAC /LDAC pin
RE1  O    DAC /WR pin
RE2  O    DAC /CS pin
RD3  -    unused
RD4  -    unused
RD5  -    unused
RD6  -    unused
RD7  -    unused
**********************************************************/

// MCU config:
//	bit 12-13= 11 - FLASH Program Memory Code Protection off
//  bit 11   = 0  - In-Circuit Debugger enabled
//	bit 10   = 1  - unimplemented (read as 1)
//	bit 9    = 0  - Unprotected program memory may not be written to by EECON control
//	bit 8    = 1  - Data EE Memory Code Protection off
//	bit 7    = 0  - RB3 is digital I/O, HV on MCLR must be used for programming
//	bit 6    = 0  - BOR disabled
//	bit 4-5  = 11 - Code protection off
//	bit 3    = 1  - PWRT disabled
//	bit 2    = 0  - WDT disabled
//	bit 0-1  = 10 - high-speed oscillator
__CONFIG(0x373A);

// Oscillator frequency
#define F_OSC					20000000

// Sound sample rate
#define SAMPLE_RATE				8000

// Pulse counter threshold to start ringing
#define RINGER_THRESHOLD		10

// IR code to switch Ringer on/off
#define IR_CODE_ONOFF			1086259455
// IR codes to select sounds
#define IR_CODE_SOUND_1			1086300255
#define IR_CODE_SOUND_2			1086283935
#define IR_CODE_SOUND_3			1086316575
#define IR_CODE_SOUND_4			1086296175
// IR codes to control volume
#define IR_CODE_VOL_UP			1086322695
#define IR_CODE_VOL_DOWN		1086290055
// IR code to enter programming mode
#define IR_CODE_PROG			1086298215

// DAC pins
#define DAC_PORT				PORTD
#define DAC_AB					RA3
#define DAC_PD					RA5
#define DAC_CLR					RC2
#define DAC_LDAC				RE0
#define DAC_WR					RE1
#define DAC_CS					RE2

// Other pins
#define RS232_EN				RA1
#define RS485_EN				RA2
#define RING					RA0
#define EEPROM_WP				RB1

// Baud Rate = FOSC / (4 * (SSPADD + 1) )
// Baud Rate = 1MHz
#define I2C_BAUDRATE_CONST		4

// EEPROM base address
#define EEPROM_ADDR_BASE		0b10100000

// Function prototypes
static void init(void);
static void init_ports(void);
static void init_peripherals(void);
static void init_dac(void);
static void init_i2c(void);
static void dac_write(char d);
static void dac_test(void);
static void write_eeprom(unsigned char index);
static void play_sound(unsigned char index);
static void play_ir_sound(void);


// Current sound index
unsigned char g_sound = 0;
// Is ringer on?
unsigned char g_ison = 1;
// Is ringer button pressed?
unsigned char g_ring = 0;
// Pulse counter
unsigned char g_pulsecount = 0;
// Volume (bits to shift sample value)
unsigned char g_vol = 0;

void main(void)
{
	init();

	for (;;)
	{
		if (g_iscode)
		{
			switch (g_rcvcode)
			{
				case IR_CODE_ONOFF:
					g_ison ^= 1;
					break;

				case IR_CODE_SOUND_1:
					g_ison = 1;
					g_sound = 0;
					break;

				case IR_CODE_SOUND_2:
					g_ison = 1;
					g_sound = 1;
					break;

				case IR_CODE_SOUND_3:
					g_ison = 1;
					g_sound = 2;
					break;

				case IR_CODE_SOUND_4:
					g_ison = 1;
					g_sound = 3;
					break;

				case IR_CODE_VOL_UP:
					if (g_vol > 0)
					{
						--g_vol;
					}
					break;

				case IR_CODE_VOL_DOWN:
					if (g_vol < 7)
					{
						++g_vol;
					}
					break;

				case IR_CODE_PROG:
					play_ir_sound();
					write_eeprom(g_sound);
					break;

				default:
					continue;
			}	

			play_ir_sound();
			g_iscode = 0;

			goto play;
		}

		if (!RING)
		{
			if (++g_pulsecount == RINGER_THRESHOLD)
			{
				g_pulsecount = 0;
				goto play;
			}	
		}

		if (g_ring)
		{
			g_ring = 0;
		}
		else
		{
			continue;
		}	
			
play:
		if (g_ison)
		{
			play_sound(g_sound);
		}
	}	
		
	// Data processing loop		
	/*for (;;)
	{
		if (receive_packet())
		{
			process_packet();
		}
	}*/	
}

static void init(void)
{
	GIE = 0;	// Disable interrupts

	init_ports();
	init_peripherals();
	init_dac();

	INTEDG = 1;	// Edge-triggered external interrupt
	INTE = 1;	// Enable external interrupt
	GIE = 1;	// Enable interrupts
}

static void init_ports(void)
{
	// Port A is digital output
	ADCON1 = 0b00000110;
	PORTA = 0;
	TRISA = 0b11000001;
	PORTB = 0;
	TRISB = 0b11000001;
	PORTC = 0;
	TRISC = 0b11011000;
	PORTD = 0;
	TRISD = 0b00000000;
	PORTE = 0;
	TRISE = 0b00000000;
}

static void init_peripherals(void)
{
	sysnet_init();
	IR_Init();
	IR_SetDefConfig();
	init_i2c();
	
	// Initialize Timer2
	// Sample rate = 20000000 / 4 / 125 / 5 = 8000
	
	// 1:1 prescaler, 1:5 postscaler, disable Timer2
	T2CON = 0b00100000;
	// Period of 125
	PR2 = 125;
	
	// Enable EEPROM write protection
	EEPROM_WP = 1;
	// Enable RS-485
	RS485_EN = 0;
	// Disable RS-232
	RS232_EN = 1;
}

static void init_dac(void)
{
	DAC_PD = 1;		// Normal power mode
	DAC_WR = 1;		// Disable DAC write
	DAC_AB = 0;		// Select DAC A
	DAC_LDAC = 0;	// DACs are updated on the rising edge of WR
	
	// Clear DAC registers
	DAC_CLR = 0;
	DAC_CLR = 1;
	
	DAC_PORT = 0;	// Clear DAC port
	DAC_CS = 0;		// Enable DAC
}

static void init_i2c(void)
{
	// initialize I2C baud rate
	SSPADD = I2C_BAUDRATE_CONST;
	
	// Slew rate control disabled for standard speed mode (100 kHz and 1 MHz)
	SSPSTAT |= SMP;
	// Input levels conform to I2C specs
	SSPSTAT &= ~CKE;
	
	// Master mode, SSP enable
	SSPCON = 0b00111000;
}	

// Writes 8-bit sample to DAC
static void dac_write(char d)
{
	DAC_PORT = d;
	DAC_WR = 0;
	DAC_WR = 1;
}	

static void dac_test(void)
{
	char i = 0;
	for (;;)
	{
		dac_write(i++);
	}
}	

// Interrupt service routine
static void interrupt
isr(void)
{
	IR_HandleInterrupt();
}

static void write_eeprom(unsigned char index)
{
	unsigned short length = 0;
	unsigned short i = 0;
	unsigned char d = 0;
	unsigned char dev_addr = (index << 1) | EEPROM_ADDR_BASE;

	// Read sound length (Big Endian)
	receive_byte((unsigned char*)&length, 0);
	receive_byte((unsigned char*)&length + 1, 0);

	// Disable EEPROM write protection
	EEPROM_WP = 0;
	
	// Write sound length (Big Endian)
	while (!eeprom_write_byte(dev_addr, 0, length & 0xFF));
	while (!eeprom_write_byte(dev_addr, 1, length >> 8));

	for (; i < length; ++i)
	{
		// Read sound sample
		receive_byte(&d, 0);
		// Write sound sample
		while (!eeprom_write_byte(dev_addr, i + 2, d));
	}

	// Enable EEPROM write protection
	EEPROM_WP = 1;
}

static void play_sound(unsigned char index)
{
	unsigned short length = 0;
	unsigned char sample = 0;
	unsigned char dev_addr = (index << 1) | EEPROM_ADDR_BASE;
	unsigned short addr = 0;

	// Read sound length (Big Endian)
	eeprom_read_random(dev_addr, addr, (unsigned char*)&length);
	eeprom_read_current(dev_addr, (unsigned char*)&length + 1);

	// Enable Timer2
	TMR2ON = 1;

	for (; addr	< length; ++addr)
	{
		eeprom_read_current(dev_addr, &sample);
		sample = (((char)sample - 0x80) >> g_vol) + 0x80;
		dac_write(sample);
		// Wait for the timer to synchronize sample rate with EEPROM read
		while (!TMR2IF);
		// Reset the timer
		TMR2IF = 0;
	}

	// Disable Timer2
	TMR2ON = 0;
}	

static void play_ir_sound(void)
{
	unsigned short i = 0;

	// Enable Timer2
	TMR2ON = 1;

	for (; i < SAMPLE_RATE / 8; ++i)
	{
		dac_write(i & 8 ? 255 : 0);
		// Wait for the timer to synchronize sample rate with EEPROM read
		while (!TMR2IF);
		// Reset the timer
		TMR2IF = 0;
	}

	for (i = 0; i < SAMPLE_RATE / 8; ++i)
	{
		// Wait for the timer to synchronize sample rate with EEPROM read
		while (!TMR2IF);
		// Reset the timer
		TMR2IF = 0;
	}

	// Disable Timer2
	TMR2ON = 0;
}	
